from __future__ import print_function, division
import numpy as np

def get_spectral_norm(L):
    if L is None:
        return 1
    elif hasattr(L, "spectral_norm"):
        return L.spectral_norm
    else: # linearized ADMM
        LTL = L.T.dot(L)
        # need spectral norm of L
        import scipy.sparse
        if scipy.sparse.issparse(L):
            if min(L.shape) <= 2:
                L2 = np.real(np.linalg.eigvals(LTL.toarray()).max())
            else:
                import scipy.sparse.linalg
                L2 = np.real(scipy.sparse.linalg.eigs(LTL, k=1, return_eigenvectors=False)[0])
        else:
            L2 = np.real(np.linalg.eigvals(LTL).max())
        return L2

class MatrixAdapter(object):
    """Matrix adapter to deal with None and per-component application.
    """
    def __init__(self, L, axis=None):
        # prevent cascade
        spec_norm = None
        while isinstance(L, MatrixAdapter):
            spec_norm = L._spec_norm
            axis = L.axis
            L = L.L
        self.L = L
        self.axis = axis
        self._spec_norm = spec_norm

    @property
    def spectral_norm(self):
        if self._spec_norm is None:
            if self.L is not None:
                self._spec_norm = get_spectral_norm(self.L)
            else:
                self._spec_norm = 1
        return self._spec_norm

    @property
    def T(self):
        if self.L is None:
            return self # NOT: self.L !!!
        # because we need to preserve axis for dot(), create a new adapter
        return MatrixAdapter(self.L.T, axis=self.axis)

    def dot(self, X):
        if self.L is None:
             # CAVEAT: This is not a copy (for performance reasons)
             # so make sure you're not binding it to another variable
             # OK for all temporary arguments X
            return X

        if self.axis is None:
            return self.L.dot(X)

        # axis=0 is not needed because it can be done with a normal matrix
        # dot product
        if self.axis == 1:
            return self.L.dot(X.reshape(-1)).reshape(X.shape[0], -1)
        raise NotImplementedError("MatrixAdapter.dot() is not useful with axis=0.\n"
                                  "Use regular matrix dot product instead!")
    def __len__(self):
        return len(self.L)

    @property
    def shape(self):
        return self.L.shape

    @property
    def size(self):
        return self.L.size

    @property
    def ndim(self):
        return self.L.ndim


class Traceback(object):
    def __init__(self):
        self._trace = []

    def __call__(self, *X, it=None):
        self._trace.append(tuple(x.copy() for x in X))

    @property
    def trace(self):
        return self._trace

    def clear(self):
        self._trace = []

class NullCallback(object):
    def __call__(self, *X, it):
        pass


class ApproximateCache(object):
    """Cache function evaluations that don't change much

    Certain calculations take a long time but have little change after a few
    iterations, so that these functions become an unnecessary time sink.
    This class allows the user to store these (and similar) values and
    only calculate them when there is a substantial change in the value.

    This method works by comparing the stored value to the last value calculated.
    If the relative difference between the two is less than `slack`/2, then the
    change is deemed to be "small" and the `stride` (number of iterations to skip)
    is increased to skip calculating the value for several iterations.
    """
    def __init__(self, func, slack=0.1, max_stride=100):
        """Constructor

        Parameters
        ----------
        func: function
            The slow function that calculates the value
        slack: float, default=0.1
            A measure of how much the value can differ before
            it needs to be recalculated
        max_stride: int, default=100
            Maximum `stride` between calculations of the value.
        """
        self.func = func
        assert slack >= 0 and slack < 1
        self.slack = slack
        self.max_stride = max_stride
        self.it = 0
        self.stride = 1
        self.last = -1
        self.stored = None

    def __len__(self):
        """The current `stride` value
        """
        return len(self.stride)

    def __call__(self, *args, **kwargs):
        """Calculate the value (if necessary)

        This method checks whether or not the value needs to be recalculated,
        and if so, calculates it and sets the number of `stride` (iterations)
        until it is calculated again.
        """
        if self.slack == 0:
            self.it += 1
            return self.func(*args, **kwargs)
        if self.it >= self.last + self.stride:
            self.last = self.it
            val = self.func(*args, **kwargs)

            # increase stride when rel. changes in L are smaller than (1-slack)/2
            if self.it > 1 and self.slack > 0:
                rel_error = np.abs(self.stored - val) / self.stored
                budget = self.slack/2
                if rel_error < budget and rel_error > 0:
                    self.stride += max(1,int(budget/rel_error * self.stride))
                    self.stride = min(self.max_stride, self.stride)
            # updated last value
            self.stored = val
        else:
            self.it += 1
        return self.stored

class NesterovStepper(object):
    def __init__(self, accelerated=False):
        self.t = 1.
        self.accelerated = accelerated

    @property
    def omega(self):
        if self.accelerated:
            t_ = 0.5*(1 + np.sqrt(4*self.t*self.t + 1))
            om = (self.t - 1)/t_
            self.t = t_
            return om
        else:
            return 0

def initZU(X, L):
    if not isinstance(L, list):
        Z = L.dot(X).copy()
        U = np.zeros(Z.shape, dtype=Z.dtype)
    else:
        Z = []
        U = []
        for i in range(len(L)):
            Z.append(L[i].dot(X).copy())
            U.append(np.zeros(Z[i].shape, dtype=Z[i].dtype))
    return Z,U

def l2sq(x):
    """Sum the matrix elements squared
    """
    return (x**2).sum()

def l2(x):
    """Square root of the sum of the matrix elements squared
    """
    return np.sqrt((x**2).sum())

def get_step_g(step_f, norm_L2, N=1, M=1):
    """Get step_g compatible with step_f (and L) for ADMM, SDMM, GLMM.
    """
    # Nominally: minimum step size is step_f * norm_L2
    # see Parikh 2013, sect. 4.4.2
    #
    # BUT: For multiple constraints, need to multiply by M.
    # AND: For multiple variables, need to multiply by N.
    # Worst case of constraints being totally correlated, otherwise Z-updates
    # overwhelm X-updates entirely -> blow-up
    return step_f * norm_L2 * N * M

def get_step_f(step_f, lR2, lS2):
    """Update the stepsize of given the primal and dual errors.

    See Boyd (2011), section 3.4.1
    """
    mu, tau = 10, 2
    if lR2 > mu*lS2:
        return step_f * tau
    elif lS2 > mu*lR2:
        return step_f / tau
    return step_f

def do_the_mm(X, step_f, Z, U, prox_g, step_g, L):
    LX = L.dot(X)
    Z_ = prox_g(LX + U, step_g)
    # primal and dual errors
    R = LX - Z_
    S = -1/step_g * L.T.dot(Z_ - Z)
    Z[:] = Z_[:] # force the copy
    # this uses relaxation parameter of 1
    U[:] += R
    return LX, R, S

def update_variables(X, Z, U, prox_f, step_f, prox_g, step_g, L):
    """Update the primal and dual variables

    Note: X, Z, U are updated inline

    Returns: LX, R, S
    """
    if not hasattr(prox_g, '__iter__'):
        if prox_g is not None:
            dX = step_f/step_g * L.T.dot(L.dot(X) - Z + U)
            X[:] = prox_f(X - dX, step_f)
            LX, R, S = do_the_mm(X, step_f, Z, U, prox_g, step_g, L)
        else:
            # fall back to simple fixed-point method for f
            # see do_the_mm for normal definitions of LX,Z,R,S
            S = -X.copy()
            X[:] = prox_f(X, step_f)
            LX = X
            Z[:] = X[:]
            R = np.zeros(X.shape, dtype=X.dtype)
            S += X

    else:
        M = len(prox_g)
        dX = np.sum([step_f/step_g[i] * L[i].T.dot(L[i].dot(X) - Z[i] + U[i]) for i in range(M)], axis=0)
        X[:] = prox_f(X - dX, step_f)
        LX = [None] * M
        R = [None] * M
        S = [None] * M
        for i in range(M):
            LX[i], R[i], S[i] = do_the_mm(X, step_f, Z[i], U[i], prox_g[i], step_g[i], L[i])
    return LX, R, S

def get_variable_errors(X, L, LX, Z, U, step_g, e_rel, e_abs=0):
    """Get the errors in a single multiplier method step

    For a given linear operator A, (and its dot product with X to save time),
    calculate the errors in the prime and dual variables, used by the
    Boyd 2011 Section 3 stopping criteria.
    """
    n = X.size
    p = Z.size
    e_pri2 = np.sqrt(p)*e_abs/L.spectral_norm + e_rel*np.max([l2(LX), l2(Z)])
    if step_g is not None:
        e_dual2 = np.sqrt(n)*e_abs/L.spectral_norm + e_rel*l2(L.T.dot(U)/step_g)
    else:
        e_dual2 = np.sqrt(n)*e_abs/L.spectral_norm + e_rel*l2(L.T.dot(U))
    return e_pri2, e_dual2

def check_constraint_convergence(X, L, LX, Z, U, R, S, step_f, step_g, e_rel, e_abs):
    """Calculate if all constraints have converged.

    Using the stopping criteria from Boyd 2011, Sec 3.3.1, calculate whether the
    variables for each constraint have converged.
    """

    if isinstance(L, list):
        M = len(L)
        convergence = True
        errors = []
        # recursive call
        for i in range(M):
            c, e = check_constraint_convergence(X, L[i], LX[i], Z[i], U[i], R[i], S[i],
                                                step_f, step_g[i], e_rel, e_abs)
            convergence &= c
            errors.append(e)
        return convergence, errors
    else:
        # check convergence of prime residual R and dual residual S
        e_pri, e_dual = get_variable_errors(X, L, LX, Z, U, step_g, e_rel, e_abs)
        lR = l2(R)
        lS = l2(S)
        convergence = (lR <= e_pri) and (lS <= e_dual)
        return convergence, (e_pri, e_dual, lR, lS)

def check_convergence(newX, oldX, e_rel):
    """Check that the algorithm converges using Langville 2014 criteria

    Uses the check from Langville 2014, Section 5, to check if the NMF
    algorithm has converged.
    """
    # Calculate the norm for columns and rows, which can be used for debugging
    # Otherwise skip, since it takes extra processing time
    new_old = newX*oldX
    old2 = oldX**2
    norms = [np.sum(new_old), np.sum(old2)]
    convergent = norms[0] >= (1-e_rel**2)*norms[1]
    return convergent, norms

def hasNotNone(l):
    i = 0
    for ll in l:
        if ll is not None:
            if hasattr(ll, '__iter__'):
                for lll in ll:
                    if lll is not None:
                        return len(l) - i
        i += 1
    return 0



import numpy as np

#### CAUTION ####
def _step_gamma(step, gamma):
    """Update gamma parameter for use inside of continuous proximal operator.

    Every proximal operator for a function with a continuous parameter,
    e.g. gamma ||x||_1, needs to update that parameter to account for the
    stepsize of the algorithm.

    Returns:
        gamma * step
    """
    return gamma * step
#################

def prox_id(X, step):
    """Identity proximal operator
    """
    return X

def prox_zero(X, step):
    """Proximal operator to project onto zero
    """
    X[:] = np.zeros(X.shape, dtype=X.dtype)
    return X

def prox_plus(X, step):
    """Projection onto non-negative numbers
    """
    below = X < 0
    X[below] = 0
    return X

def prox_unity(X, step, axis=0):
    """Projection onto sum=1 along an axis
    """
    X[:] = X / np.sum(X, axis=axis, keepdims=True)
    return X

def prox_unity_plus(X, step, axis=0):
    """Non-negative projection onto sum=1 along an axis
    """
    X[:] = prox_unity(prox_plus(X, step), step, axis=axis)
    return X

def prox_min(X, step, thresh=0):
    """Projection onto numbers above `thresh`
    """
    thresh_ = _step_gamma(step, thresh)
    below = X - thresh_ < 0
    X[below] = thresh_
    return X

def prox_max(X, step, thresh=0):
    """Projection onto numbers below `thresh`
    """
    thresh_ = _step_gamma(step, thresh)
    above = X - thresh_ > 0
    X[above] = thresh_
    return X

def prox_components(X, step, prox=None, axis=0):
    """Split X along axis and apply prox to each chunk.

    prox can be a list.
    """
    K = X.shape[axis]

    if not hasattr(prox_list, '__iter__'):
        prox = [prox] * K
    assert len(prox_list) == K

    if axis == 0:
        Pk = [prox_list[k](X[k], step) for k in range(K)]
    if axis == 1:
        Pk = [prox_list[k](X[:,k], step) for k in range(K)]
    X[:] = np.stack(Pk, axis=axis)
    return X


#### Regularization function below ####

def prox_hard(X, step, thresh=0):
    """Hard thresholding

    X if |X| >= thresh, otherwise 0
    NOTE: modifies X in place
    """
    thresh_ = _step_gamma(step, thresh)
    below = np.abs(X) < thresh_
    X[below] = 0
    return X

def prox_hard_plus(X, step, thresh=0):
    """Hard thresholding with projection onto non-negative numbers
    """
    X[:] = prox_plus(prox_hard(X, step, thresh=thresh), step)
    return X

def prox_soft(X, step, thresh=0):
    """Soft thresholding proximal operator
    """
    thresh_ = _step_gamma(step, thresh)
    X[:] = np.sign(X)*prox_plus(np.abs(X) - thresh_, step)
    return X

def prox_soft_plus(X, step, thresh=0):
    """Soft thresholding with projection onto non-negative numbers
    """
    X[:] = prox_plus(prox_soft(X, step, thresh=thresh), step)
    return X

def prox_max_entropy(X, step, gamma=1):
    """Proximal operator for maximum entropy regularization.

    g(x) = gamma \sum_i x_i ln(x_i)

    has the analytical solution of gamma W(1/gamma exp((X-gamma)/gamma)), where
    W is the Lambert W function.
    """
    from scipy.special import lambertw
    gamma_ = _step_gamma(step, gamma)
    # minimize entropy: return gamma_ * np.real(lambertw(np.exp((X - gamma_) / gamma_) / gamma_))
    above = X > 0
    X[above] = gamma_ * np.real(lambertw(np.exp(X[above]/gamma_ - 1) / gamma_))
    return X

class AlternatingProjections(object):
    """Combine several proximal operators in the form of Alternating Projections

    This implements the simple POCS method with several repeated executions of
    the projection sequence.

    Note: The operators are executed in the "natural" order, i.e. the first one
    in the list is applied last.
    """
    def __init__(self, prox_list=None, repeat=1):
        self.operators = []
        self.repeat = repeat
        if prox_list is not None:
            self.operators += prox_list

    def __call__(self, X, step):
        # simple POCS method, no Dykstra or averaging
        # TODO: no convergence test
        # NOTE: inline updates
        for r in range(self.repeat):
            # in reverse order (first one last, as expected from a sequence of ops)
            for prox in self.operators[::-1]:
                X = prox(X, step)
        return X

    def find(self, cls):
        import functools
        for i in range(len(self.operators)):
            prox = self.operators[i]
            if isinstance(prox, functools.partial):
                if prox.func is cls:
                    return i
            else:
                if prox is cls:
                    return i
        return -1

def get_gradient_y(shape, py):
    """Calculate the gradient in the y direction to the line at py

    The y gradient operator is a block matrix, where each block is the size of the image width.
    The matrix itself is made up of (img_height x img_height) blocks, most of which are all zeros.
    """
    import scipy.sparse

    height, width = shape
    rows = []
    empty = scipy.sparse.dia_matrix((width, width))
    identity = scipy.sparse.identity(width)

    # Create the blocks by row, beginning with blocks leading up to the peak row from the top
    for n in range(py):
        row = [empty]*n
        row += [-identity, identity]
        row += [empty]*(height-n-2)
        rows.append(row)
    # Set all elements in the peak row to zero
    rows.append([empty]*height)
    # Create the blocks for the rows leading up to the peak row from the bottom
    for n in range(height-py-1):
        row = [empty]*(py+n)
        row += [identity, -identity]
        row += [empty]*(height-py-n-2)
        rows.append(row)
    return scipy.sparse.bmat(rows)

def get_gradient_x(shape, px):
    """Calculate the gradient in the x direction to the line at px

    The y gradient operator is a block diagonal matrix, where each block is the size of the image width.
    The matrix itself is made up of (img_height x img_height) blocks, most of which are all zeros.
    """
    import scipy.sparse

    height, width = shape
    size = height * width

    # Set the diagonal to -1, except for the value at the peak, which is zero
    c = -np.ones((width,))
    c[px] = 0
    # Set the pixels leading up to the peak from the left
    r = np.zeros(c.shape, dtype=c.dtype)
    r[:px] = 1
    # Set the pixels leading up to the peak from the right
    l = np.zeros(c.shape, dtype=c.dtype)
    l[px:] = 1
    # Make a block for a single row in the image
    block = scipy.sparse.diags([l, c, r], [-1, 0,1], shape=(width,width))
    # Use the same block for each row
    op = scipy.sparse.block_diag([block for n in range(height)])
    return op



from __future__ import print_function, division
import numpy as np
import scipy.sparse, scipy.sparse.linalg
from . import operators
from . import utils
from . import algorithms

import logging
logger = logging.getLogger("proxmin")

def log_likelihood(*X, Y=0, W=1):
    """Log-likelihood of NMF assuming Gaussian error model.

    Args:
        X:  tuple of (A,S) matrix factors
        Y:  target matrix
        W: (optional weight matrix MxN)

    Returns:
        float
    """
    A, S = X
    return np.sum(W*(Y - A.dot(S))**2) / 2

def grad_likelihood(*X, Y=0, W=1):
    """Gradient of the log-likelihood of NMF assuming Gaussian error model.

    Args:
        X:  tuple of (A,S) matrix factors
        Y:  target matrix
        W: (optional weight matrix MxN)

    Returns:
        grad_A f, grad_S f
    """
    A, S = X
    D = W*(A.dot(S) - Y)
    return D.dot(S.T), A.T.dot(D)

def step_A(A, S):
    return 1/utils.get_spectral_norm(S.T)

def step_S(A, S):
    return 1/utils.get_spectral_norm(A)

def step_pgm(*X, it=None, W=1):
    """Step sizes for PGM of NMF assuming Gaussian error model.

    Args:
        X:  tuple of (A,S) matrix factors
        it:  iteration counter
        W: (optional weight matrix MxN)

    Returns:
        step_A, step_S
    """
    A, S = X
    if W is 1:
        return step_A(A,S), step_S(A,S)
    else:
        C, K = A.shape
        K, N = S.shape
        Sigma_1 = scipy.sparse.diags(W.flatten())

        # Lipschitz constant for grad_A = || S Sigma_1 S.T||_s
        PS = scipy.sparse.block_diag([S.T for c in range(C)])
        SSigma_1S = PS.T.dot(Sigma_1.dot(PS))
        LA = np.real(scipy.sparse.linalg.eigs(SSigma_1S, k=1, return_eigenvectors=False)[0])

        # Lipschitz constant for grad_S = || A.T Sigma_1 A||_s
        PA = scipy.sparse.bmat([[scipy.sparse.identity(N) * A[c,k] for k in range(K)] for c in range(C)])
        ASigma_1A = PA.T.dot(Sigma_1.dot(PA))
        LS = np.real(scipy.sparse.linalg.eigs(ASigma_1A, k=1, return_eigenvectors=False)[0])
        LA, LS

        return 1/LA, 1/LS

def step_adaprox(*X, it=None):
    A, S = X
    return (np.mean(A, axis=0) / 10, S.mean(axis=1)[:,None] / 10)


def nmf(Y, A, S, W=1, prox_A=operators.prox_plus, prox_S=operators.prox_plus, algorithm=algorithms.pgm, step=None, max_iter=1000, e_rel=1e-3, callback=None, **algorithm_args):
    """Non-negative matrix factorization.

    This method solves the matrix factorization problem
        minimize || Y - AS ||_2^2
    under an arbitrary number of constraints on A and/or S.

    Args:
        Y: target matrix MxN
        A: initial amplitude matrix MxK, will be updated
        S: initial source matrix KxN, will be updated
        W: (optional weight matrix MxN)
        prox_A: direct contraint of A
        prox_S: direct constraint of S
        algorithm: an algorithm from `proxmin.algorithms`
        step: function to compute step sizes for A and S
            Signature: step(*X, it)
        max_iter: maximum iteration number, irrespective of current residuals
        e_rel: relative error threshold for primal and dual residuals
        callback: arbitrary logging function
            Signature: callback(*X, it=None)
        algorithm_args: further arguments passed to the algorithm

    Returns:
        return arguments of algorithm
        A, S are updated inline

    Reference:
        Moolekamp & Melchior, 2018 (arXiv:1708.09066)

    """

    assert algorithm in [algorithms.pgm, algorithms.adaprox, algorithms.bsdmm]

    # setup
    from functools import partial
    grad = partial(grad_likelihood, Y=Y, W=W)
    X = [A, S]
    prox = [prox_A, prox_S]

    if algorithm is algorithms.pgm:
        if step is None:
            step = partial(step_pgm, W=W)
        return algorithm(X, grad, step, prox=prox, max_iter=max_iter, e_rel=e_rel, callback=callback, **algorithm_args)

    if algorithm is algorithms.adaprox:
        if step is None:
            step = step_adaprox
        return algorithm(X, grad, step, prox=prox, max_iter=max_iter, e_rel=e_rel, callback=callback, **algorithm_args)

    if algorithm is algorithms.bsdmm:

        # transform gradient steps into prox
        def prox_f(X, step, Xs=None, j=None):
            # that's a bit of a waste since we compute all gradients
            grads = grad(*Xs)
            # ...but only use one
            return prox[j](X - step * grads[j], step)


        if step is None:
            step_ = partial(step_pgm, W=W)
            def step_f(Xs, j=None):
                return step_(*Xs)[j]
            step = step_f

        return algorithms.bsdmm(X, prox_f, step_f, max_iter=max_iter, e_rel=e_rel, callback=callback, **algorithm_args)
